#!/usr/bin/env python
# coding: utf-8

from .cat_target_drift_tab import CatTargetDriftTab
from .classification_performance_tab import ClassificationPerformanceTab
from .data_drift_tab import DataDriftTab
from .data_quality_tab import DataQualityTab
from .num_target_drift_tab import NumTargetDriftTab
from .prob_classification_performance_tab import ProbClassificationPerformanceTab
from .regression_performance_tab import RegressionPerformanceTab
